/*
 *  Program to compute shading and shadows for terrain data.
 *
 *
 */
#include "matrix.h"
#include "shade.h"
#include "math.h"
#include <stdio.h>

double EastEnd;
double WestEnd;
double NorthPole;
double SouthPole;
double MtHimalaya =0;
double DeathValley;
double angle = M_PI/4;    /* Later this can be read in */
double red_diff = 0.1;
double blue_diff = 0.1;
double green_diff = 0.1;
double red_amb = 0.2;
double blue_amb = 0.2;
double green_amb = 0.2;
double epsilon = 0.1;

/* opens filename and creates data structure */
#ifdef ANSI
void readInData(char* fname)
#else
void readInData(fname)
     char* fname;
#endif
{
  FILE* infile;
  int i,j;

  if (!(infile=fopen(fname, "r")))
    myerror(1);
  
  fscanf(infile, "%d", &nverts);
  vlist=(Poly_vert*)malloc(nverts*sizeof(Poly_vert));
  if (!vlist)
    myerror(3);
  for(i=0; i<nverts; i++) 
    {
      fscanf(infile,"%lf %lf %lf", &vlist[i].x, &vlist[i].y, &vlist[i].z);
      /**** Compensate so triangles are big enough *******/
      vlist[i].x *=10;
      vlist[i].y *=1;
      vlist[i].z *=1;
      
      
      /******** Find Landmarks *********/
      if (i==0 || (vlist[i].x > EastEnd))
	EastEnd = vlist[i].x;
      if (i==0 || (vlist[i].x < WestEnd))
	WestEnd = vlist[i].x;
      if (i==0 || vlist[i].z > NorthPole)
	NorthPole = vlist[i].z;
      if (i==0 || vlist[i].z < SouthPole)
	SouthPole = vlist[i].z;
      if (vlist[i].y > MtHimalaya)
	MtHimalaya  = vlist[i].y;
      if (i==0 || vlist[i].y < DeathValley)
	  DeathValley = vlist[i].y;
      /******* Define initial colors ******/
      if (vlist[i].y == 0)         /* Blue Water */
	{
	  vlist[i].r = 0.0;
	  vlist[i].g = 0.0;
	  vlist[i].b = 1.0;
	}
      else                        /* Green Land */
	{
	  vlist[i].r = 0.0;
	  vlist[i].g = 1.0;
	  vlist[i].b = 0.0;
	}
      for(j=0; j<4; j++)           /* initialize */
	vlist[i].normal[j] = 0.0;
    }

  /** center vertices **/
  for (i = 0; i<nverts; i++)
    {
      vlist[i].x -= WestEnd;
      vlist[i].z -= SouthPole;
      vlist[i].y -= MtHimalaya;
    }
  

  fscanf(infile, "%d", &npolys);
  plist =(Triangle*)malloc(npolys*sizeof(Triangle));
  if (!plist)
    myerror(3);
  for (i=0; i<npolys; i++)
    { 
      fscanf(infile,"%d %d %d", &plist[i].v[0], &plist[i].v[1], &plist[i].v[2]);
      plist[i].n =3;
      plist[i].degenerate = FALSE;
    }

  fclose(infile);

}

#ifdef ANSI 
void copyPoint(point* p, Poly_vert* v)
#else
     void copyPoint(p,v)
     point* p;
     Poly_vert* v;
#endif
{
  p->x = v->x;
  p->y = v->y;
  p->z = v->z;
}
#ifdef ANSI
void averageIn(vector n, Poly_vert* v)
#else
     averageIn(n,v)
     vector n;
     Poly_vert* v;
#endif    
{
  int i;
  for(i=0; i<3; i++)
    {
      v->normal[i]+=n[i];
    }
  /* unit_vector(v->normal, v->normal); */
 
}

/*** Returns True if same, FALSE is different ****/
#ifdef ANSI
int cmpPoints(point* p0, point* p1)
#else
     int cmpPoints(p0, p1)
     point *p0, *p1;
#endif
{
  return ((p0->x == p1->x) && (p0->y == p1->y) && (p0->z == p1->z));
}

void computeNormals()
{
  int i,j;
  vector n, a, b;
  int v1, v2, v0;
  point p1, p2, p0;
  for (i=0; i<npolys; i++)
    {
      v0 = plist[i].v[0];
      v1 = plist[i].v[1];
      v2 = plist[i].v[2];
      copyPoint(&p0, &(vlist[v0]));
      copyPoint(&p1, &(vlist[v1]));
      copyPoint(&p2, &(vlist[v2]));
      if (cmpPoints(&p0, &p1) || cmpPoints(&p1, &p2) || 
	  cmpPoints(&p0, &p2))
	{
	  plist[i].degenerate = TRUE;
	  printf("Degenerate polygon %d\n", i);
	}
      else
	{
	  unit_line_vector(a, &p1, &p0);
	  unit_line_vector(b, &p2, &p0);
	  cross_product(n, a, b);
	  averageIn(n, &vlist[v0]);
	  averageIn(n, &vlist[v1]);
	  averageIn(n, &vlist[v2]);
	  for (j=0; j<4; j++)
	    {
	      plist[i].normal[j] = n[j];
	    }
	}
    }
  for (i=0; i<nverts; i++)
    unit_vector(vlist[i].normal, vlist[i].normal);
}

#ifdef ANSI
vert2vector(vector vect, Poly_vert* vert)
#else 
     vert2vector(vect, vert)
     vector vect;
     Poly_vert* vert;
#endif
{
  vect[EAST] = vert->x;
  vect[UP] = vert->y;
  vect[NORTH] = vert->z;
  vect[3] = 1.0;
}
#ifdef ANSI
double max(double x, double y)
#else
     double max(x, y)
     double x;
     double y;
#endif
{
  return ((x > y)? x : y);
}

void castShadows()
{
  int i,j;
  matrix transform;
  vector real, sky, move;
  int my_x, my_z;
  int yvertex;
  vector temp;
  Window win;
  vector topright, botleft, botright;
  double maxsize, altitude;

  /**** transform set of vertices *****/
  id_matrix(transform);
  rotate_z(transform, angle);

  
  for (i=0;i< nverts; i++)
    {
      vert2vector(real, &vlist[i]);
      vector_transform(sky, real, transform);
      vlist[i].sx = sky[EAST];
      vlist[i].sy = sky[UP];
      vlist[i].sz = sky[NORTH];
      vlist[i].sw = sky[3];
    }

  /**** Have to transform EastEnd and WestEnd to get right size!!!!*****/
  altitude = DeathValley - MtHimalaya;

  double2vector(topright, (EastEnd-WestEnd), 0, 0);
  double2vector(botright, (EastEnd-WestEnd), altitude, 0);
  double2vector(botleft, 0, altitude, 0);
  
  vector_transform(topright, topright, transform);
  vector_transform(botright, botright, transform);
  vector_transform(botleft, botleft, transform);

  maxsize = max(botleft[EAST],  max(topright[EAST], botright[EAST]));
  xsize = (int)ceil(maxsize) +1;
  zsize = (int)ceil(NorthPole - SouthPole)+1;


  /** Set up Y buffer **/
  Ybuffer = (Y_struct**)malloc(xsize*sizeof(Y_struct*));
  if (!Ybuffer)
    myerror(3);
  for (i = 0; i<xsize; i++)
    {
      Ybuffer[i] = (Y_struct*)malloc(zsize*sizeof(Y_struct));
      if (!Ybuffer)
	myerror(3);
      for (j=0; j<zsize; j++)
	{

	  Ybuffer[i][j].y = -1*HUGE; /* A really small number*/
	  Ybuffer[i][j].vertex = -1;
	}
    }
  win.x0 = 0;
  win.x1 = xsize -1;
  win.z0 = 0;
  win.z1 = zsize-1;

  /** Scan each polygon with back face culling **/
  for (i=0; i<npolys; i++)
    {
      if (!plist[i].degenerate)
	{
	  vector_transform(temp, plist[i].normal, transform);
	  if (temp[UP] >= 0)
	    poly_scan(plist[i], win);
	  else
	    printf("Back polygon %d\n", i);
	}
    }
  /** Scan each vertex **/
  for (i=0; i<nverts; i++)
    {
      my_x =ceil(vlist[i].sx -0.5);
      my_z =ceil(vlist[i].sz-0.5);
      if (Ybuffer[my_x][my_z].y <= (vlist[i].sy+epsilon))
	{
	  Ybuffer[my_x][my_z].vertex = i;
	  Ybuffer[my_x][my_z].y = vlist[i].sy;
	}
    }
}

#ifdef ANSI
void setpoint(point* p, double x, double y, double z)
#else
     setpoint(p, x, y,z)
     point* p;
     double x, y, z;
#endif
{
  p->x = x;
  p->y = y;
  p->z = z;
}

void computeIntensities()
{
  double *red_ints; /* Temporary intensity for each vertex */
  double *blue_ints;
  double *green_ints;
  double dot;
  int vertex, i,j;
  point p0, p1, p2;
  vector a, b, lightnormal;

  red_ints = (double*)malloc(nverts*sizeof(double));
  blue_ints = (double*)malloc(nverts*sizeof(double));
  green_ints = (double*)malloc(nverts*sizeof(double));
  
  /*** Find Normal of light ***/
  setpoint(&p0, 0.0, 0.0, 0.0);
  setpoint(&p1, cos(angle), sin(angle), 0.0);
  unit_line_vector(lightnormal, p1, p0);
  
  for (i = 0; i<xsize; i++)
    for (j = 0; j<zsize; j++)
      {
	vertex = Ybuffer[i][j].vertex;
	if (vertex != -1)
	  {
	    /* printf("I can see vertex %d at %d %d\n", vertex, i, j); */
	    dot_product(&dot, lightnormal, vlist[vertex].normal);
	    red_ints[vertex] = red_diff * dot;
	    blue_ints[vertex] = blue_diff * dot;
	    green_ints[vertex] = green_diff * dot;
	  }
      }
  for (i = 0; i<nverts; i++)
    {
      red_ints[i] += red_amb;
      blue_ints[i] += blue_amb;
      green_ints[i] += green_amb;
      
      if (red_ints[i] > 1.0)
	red_ints[i] =1.0;
      if (blue_ints[i] >1.0)
	blue_ints[i] = 1.0;
      if (green_ints[i] > 1.0)
	green_ints[i] = 1.0;
      
      vlist[i].r *= red_ints[i];
      vlist[i].g *= green_ints[i];
      vlist[i].b *= blue_ints[i];
    }
  free(red_ints);
  free(blue_ints);
  free(green_ints);
}

      
void printResults()
{
  FILE *out;
  int i;
  out = fopen("rgb.out", "w");
  if (!out)
    myerror(0);
  for (i =0; i<nverts; i++)
      fprintf(out, "%lf %lf %lf \n", vlist[i].r, vlist[i].g, vlist[i].b);
  fflush(out);
  fclose(out);
}
      
	    



#ifdef ANSI
main(int argc, char** argv)
#else
     main(argc, argv)
     int argc;
     char** argv;
#endif
{
  
  if (argc < 2)
    myerror(2);

  readInData(argv[1]);
  computeNormals();
  castShadows();
  computeIntensities();
  printResults();
}













